#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "swapper.h"
#include "swap_strategies.h"
#include "swap_helpers.h"

/**
 * Defines the memory access pattern, each pair is a combination 
 * of the accessed virtual page and either READ or WRITE
 */
struct page_access access_list[] = {
	{ 0, READ_ACCESS},
	{ 3, WRITE_ACCESS},
	{ 1, READ_ACCESS},
	{ 2, WRITE_ACCESS}, 
	{ 0, WRITE_ACCESS},
	{ 4, READ_ACCESS},
	{ 0, READ_ACCESS},
	{ 3, READ_ACCESS},
	{ 1, READ_ACCESS},
};
#define ACCESS_COUNT sizeof(access_list)/sizeof(struct page_access)

/** 
 * Management information for our memory simulation. We have an entry for 
 * each virtual page that is in our address space and an entry for each physical
 * frame that is built into the system.
 *
 * At all we have NUMBER_VIRTUAL_PAGES virtual pages and NUMBER_PHYSICAL_FRAMES
 * physical frames in our system. All structs are defined in swapper.h
 * 
 * All physcial frames are unassigned to virtual pages in the beginning.
 * Whenever there is a memory access, a new physical frame is searched if 
 * necessary.
 */
struct virtual_page_entry virtual_page_table[NUMBER_VIRTUAL_PAGES];
struct physical_frame_entry physical_frames[NUMBER_PHYSICAL_FRAMES];

/**
 * The following defines our swapping strategy. The swapping strategies 
 * are defined in swap_strategies.{c,h}. Uncomment the strategy that you 
 * want to test (and comment out the others).
 */
//struct physical_frame_entry* (*swap_strategy)(struct physical_frame_entry*) = &stupid_swapper;
struct physical_frame_entry* (*swap_strategy)(struct physical_frame_entry*) = &extended_clock_swap_strategy;

/**
 * Method maps a virtual page to a phyiscal frame
 */
void map_page_frame(struct virtual_page_entry* virtual_page, struct physical_frame_entry* physical_frame) {
	SET_BIT(virtual_page->bits, PRESENT_BIT);
	virtual_page->mapped_physical_frame = physical_frame;
	physical_frame->virtual_page = virtual_page;
}

/**
 * Method undoes a previous mapping and clears all relevant management
 * information from the phyiscal page entry and the virtual page entry.
 */
void unmap_physical_frame(struct physical_frame_entry* physical_frame) {
	struct virtual_page_entry* virtual_page = physical_frame->virtual_page;
	CLEAR_BIT(virtual_page->bits, PRESENT_BIT);
	CLEAR_BIT(virtual_page->bits, R_BIT);
	CLEAR_BIT(virtual_page->bits, M_BIT);
	virtual_page->mapped_physical_frame = NULL;
	physical_frame->virtual_page = NULL;
}

/** 
 * This method performs all steps necessary to simulate a memory access. 
 *
 * It identifies if there is physical frame assigned to the virtual page, 
 * selects a new frame if there is none and updates the R-BIT, M-BIT and 
 * PRESENT bits. 
 *
 * If no more physical memory is available, it will call the swapping strategy 
 * to determine which physical frame should be swapped and reused for our 
 * current memory access.
 */ 
void perform_memory_access(int virtual_page_number, int access) {
	struct virtual_page_entry* virtual_page = virtual_page_table + virtual_page_number;
	// check if the page is available and mapped to the virtual page
	if (!IS_BIT_SET(virtual_page->bits, PRESENT_BIT)) {
		// page is not mapped. Try to get an yet unused physical page
		// and map the virtual page onto this physical page
		struct physical_frame_entry* phys_entry = NULL;
		for (int i = 0; i != NUMBER_PHYSICAL_FRAMES; ++i) {
			if (physical_frames[i].virtual_page == NULL) {
				// great we found a physical frame
				phys_entry = physical_frames + i;
				break;
			}
		}

		// check if we have identified an unused physical frame and 
		// identify a candidate for swapping otherwise
		if (phys_entry == NULL) {
			phys_entry = swap_strategy(physical_frames);
			if (phys_entry == NULL) {
				fprintf(stderr, "KERNEL PANIC: swap_strategy did not provide a physical frame to be swapped. Cannot continue.\n");
				exit(1);
			}
			unmap_physical_frame(phys_entry);
		}
		map_page_frame(virtual_page, phys_entry);
	}
	// update bits 
	if (access == READ_ACCESS)
		SET_BIT(virtual_page->bits, R_BIT);
	if (access == WRITE_ACCESS) {
		SET_BIT(virtual_page->bits, R_BIT);
		SET_BIT(virtual_page->bits, M_BIT);
	}
}

int main(int argc, char** argv) {
	init_structs();	
	print_table_header();
	for (int i = 0; i != ACCESS_COUNT; ++i) {
		// perform the access to the memory location
		perform_memory_access(access_list[i].virtual_page_number, access_list[i].access);
		// write the page assignment and flags after the access
		print_page_assignment(access_list[i]);
	}
}
